#include "SPIFFSEditor.h"
#include <FS.h>
#define SPIFFS_MAXLENGTH_FILEPATH 32

typedef struct ExcludeListS {
    char *item;
    ExcludeListS *next;
} ExcludeList;

static ExcludeList *excludes = NULL;

static bool matchWild(const char *pattern, const char *testee) {
  const char *nxPat = NULL, *nxTst = NULL;

  while (*testee) {
    if (( *pattern == '?' ) || (*pattern == *testee)){
      pattern++;testee++;
      continue;
    }
    if (*pattern=='*'){
      nxPat=pattern++; nxTst=testee;
      continue;
    }
    if (nxPat){ 
      pattern = nxPat+1; testee=++nxTst;
      continue;
    }
    return false;
  }
  while (*pattern=='*'){pattern++;}  
  return (*pattern == 0);
}

static bool addExclude(const char *item){
    size_t len = strlen(item);
    if(!len){
        return false;
    }
    ExcludeList *e = (ExcludeList *)malloc(sizeof(ExcludeList));
    if(!e){
        return false;
    }
    e->item = (char *)malloc(len+1);
    if(!e->item){
        free(e);
        return false;
    }
    memcpy(e->item, item, len+1);
    e->next = excludes;
    excludes = e;
    return true;
}

static void loadExcludeList(fs::FS &_fs, const char *filename){
    static char linebuf[SPIFFS_MAXLENGTH_FILEPATH];
    fs::File excludeFile=_fs.open(filename, "r");
    if(!excludeFile){
        return;
    }
    if(excludeFile.isDirectory()){
      excludeFile.close();
      return;
    }
    if (excludeFile.size() > 0){
      uint8_t idx;
      bool isOverflowed = false;
      while (excludeFile.available()){
        linebuf[0] = '\0';
        idx = 0;
        int lastChar;
        do {
          lastChar = excludeFile.read();
          if(lastChar != '\r'){
            linebuf[idx++] = (char) lastChar;
          }
        } while ((lastChar >= 0) && (lastChar != '\n') && (idx < SPIFFS_MAXLENGTH_FILEPATH));

        if(isOverflowed){
          isOverflowed = (lastChar != '\n');
          continue;
        }
        isOverflowed = (idx >= SPIFFS_MAXLENGTH_FILEPATH);
        linebuf[idx-1] = '\0';
        if(!addExclude(linebuf)){
            excludeFile.close();
            return;
        }
      }
    }
    excludeFile.close();
}

static bool isExcluded(fs::FS &_fs, const char *filename) {
  ExcludeList *e = excludes;
  while(e){
    if (matchWild(e->item, filename)){
      return true;
    }
    e = e->next;
  }
  return false;
}

SPIFFSEditor::SPIFFSEditor(const fs::FS& fs)
:_fs(fs)
{}

bool SPIFFSEditor::canHandle(AsyncWebServerRequest *request){
  if(request->url().equalsIgnoreCase("/edit")){
    if(request->method() == HTTP_GET){
      if(request->hasParam("list"))
        return true;
      if(request->hasParam("edit")){
		String filename = request->arg("edit");
		if(!filename.startsWith("/")) filename = "/"+filename;
        //request->_tempFile = _fs.open(request->arg("edit"), "r");
		request->_tempFile = _fs.open(filename, "r");
		Serial.print("(1)canHandle->HTTP_GET->hasParam(\"edit\"):  "); Serial.print("("); Serial.print(request->arg("edit")); Serial.print(")"); Serial.println(filename); // 
        if(!request->_tempFile){
          return false;
        }
        if(request->_tempFile.isDirectory()){
          request->_tempFile.close();
          return false;
        }
      }
      if(request->hasParam("download")){
		String filename = request->arg("download");
		if(!filename.startsWith("/")) filename = "/"+filename;
        //request->_tempFile = _fs.open(request->arg("download"), "r");
		request->_tempFile = _fs.open(filename, "r");
		Serial.print("(1)canHandle->HTTP_GET->hasParam(\"download\"):  "); Serial.print("("); Serial.print(request->arg("download")); Serial.print(")"); Serial.println(filename); // 
        if(!request->_tempFile){
          return false;
        }
        if(request->_tempFile.isDirectory()){
          request->_tempFile.close();
          return false;
        }
      }
      request->addInterestingHeader("If-Modified-Since");
      return true;
    }
    else if(request->method() == HTTP_POST)
      return true;
    else if(request->method() == HTTP_DELETE)
      return true;
    else if(request->method() == HTTP_PUT)
      return true;
  }
  return false;
}

void SPIFFSEditor::handleRequest(AsyncWebServerRequest *request){
  
  if(request->method() == HTTP_GET){
    if(request->hasParam("list")){
      String path = request->getParam("list")->value();
      File dir = _fs.open(path);
      path = String();
      String output = "[";
      File entry = dir.openNextFile();
      while(entry){
        if (isExcluded(_fs, entry.name())) {
            entry = dir.openNextFile();
            continue;
        }
        if (output != "[") output += ',';
        output += "{\"type\":\"";
        output += "file";
        output += "\",\"name\":\"";
		output += String(entry.name()).substring(1);
        //output += String(entry.name());
        output += "\"";
        output += "}";
        entry = dir.openNextFile();
      }
      dir.close();
      output += "]";
      request->send(200, "text/json", output);
	  Serial.print("(2)handleRequest->HTTP_GET->request->send(output):  "); Serial.println(output); // 
      output = String();
    }
    else if(request->hasParam("edit") || request->hasParam("download")){
	  Serial.print("(2)handleRequest->HTTP_GET->hasParam(\"edit\"&\"download\"):  "); Serial.println(request->_tempFile.name()); // 
      request->send(request->_tempFile, request->_tempFile.name(), String(), request->hasParam("download"));
    }
    else {
      //AsyncWebServerResponse *response = request->beginResponse_P(200, "text/html", edit_htm_gz, edit_htm_gz_len);
      //response->addHeader("Content-Encoding", "gzip");
      //request->send(response);
      request->send(_fs, "/edit.htm");
    }
  } else if(request->method() == HTTP_DELETE){
    if(request->hasParam("path", true)){
	  String filename = request->getParam("path", true)->value();
	  if(!filename.startsWith("/")) filename = "/"+filename;
	  _fs.remove(filename);
      //_fs.remove(request->getParam("path", true)->value());
	  Serial.print("(2)handleRequest->HTTP_DELETE->(request->getParam(\"path\", true)->value()):  "); Serial.println(request->getParam("path", true)->value()); // 
	  request->send(200, "", "DELETE: "+filename);
      //request->send(200, "", "DELETE: "+request->getParam("path", true)->value());
    } else
      request->send(404);
  } else if(request->method() == HTTP_POST){
	Serial.print("(2.1*)handleRequest->HTTP_POST->(request->getParam(\"data\", true, true):  "); Serial.println(request->hasParam("data", true, true)); // 
	Serial.print("(2.2*)handleRequest->HTTP_POST->(request->getParam(\"data\", true, true)->value()):  "); Serial.println(request->getParam("data", true, true)->value()); // 
    //if(request->hasParam("data", true, true) && _fs.exists(request->getParam("data", true, true)->value())) {
	if(request->hasParam("data", true, true)) {
	  Serial.print("(2*)handleRequest->HTTP_POST->(request->getParam(\"data\", true, true)->value()):  "); Serial.print(request->hasParam("data", true, true)); Serial.println(request->getParam("data", true, true)->value()); // 
	  //request->send(200, "", "UPLOADED: "+filename);
      request->send(200, "", "UPLOADED: "+request->getParam("data", true, true)->value());
    } else {
      request->send(500);
	  Serial.println("(2*)handleRequest->HTTP_POST-> ERROR 500"); // 
	}
  } else if(request->method() == HTTP_PUT){
    if(request->hasParam("path", true)){
      String filename = request->getParam("path", true)->value();
	  if(!filename.startsWith("/")) filename = "/"+filename;
	  Serial.print("(2)handleRequest->HTTP_PUT->(request->getParam(\"path\", true)->value()):  "); Serial.println(filename); // 
      if(_fs.exists(filename)){
        request->send(200);
      } else {
        fs::File f = _fs.open(filename, "w");
        if(f){
          f.write((uint8_t)0x00);
          f.close();
          request->send(200, "", "CREATE: "+filename);
        } else {
          request->send(500);
        }
      }
    } else
      request->send(400);
  }
}

void SPIFFSEditor::handleUpload(AsyncWebServerRequest *request, const String& filename, size_t index, uint8_t *data, size_t len, bool final){
  if(!index){
	String path = filename;
	if(!path.startsWith("/")) path = "/"+path;
    request->_tempFile = _fs.open(path, "w");
	Serial.print("(3)handleUpload:  "); Serial.println(path); // 
  }
  if(request->_tempFile){
    if(len){
      request->_tempFile.write(data,len);
    }
    if(final){
      request->_tempFile.close();
    }
  }
}
